from dataclasses import Field, dataclass, field, make_dataclass
from typing import Iterable, Optional, TypeVar, get_args

from starlette.applications import Starlette
from starlette.routing import Route
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers import UserXmlParser


class BaseService:
    registry = {}

    def __init__(self, location=None):
        self.location = location
        
    def get_asgi_app(self):
        from teserak.w3c.soap.handler import SoapHandler

        class Endpoint(SoapHandler):
            service = self

        app = Starlette(routes=[
            Route('/', Endpoint),
        ])
        return app

    def action(self, action_name):
        def inner(func):
            class Action(SoapService):
                soap_action = action_name
                soap_request_type = func.__annotations__["request"]
                input_request = get_args(soap_request_type)[0]
                output_response = func.__annotations__["return"]
                handler = func

            self.register_action(Action)

        return inner

    def register_action(self, action):
        self.registry[action.soap_action] = action

    def get_action(self, action):
        return self.registry[action]

    def get_input(self, action_cls):
        return self.registry[action_cls.soap_action].input


def soap_output_factory(field_name, soap_namespace, cls_name, cls):
    # return CalculatorSoapAddOutput
    return soap_input_factory(field_name, soap_namespace, cls_name, cls)


class Envelope:
    class Meta:
        name = "Envelope"
        namespace = "http://www.w3.org/2003/05/soap-envelope"


def soap_body_factory(field_name, cls):
    input_field = (
        field_name,
        Optional[cls],
        field(
            default=None,
            metadata={
                "name": cls.__name__,
                "type": "Element",
                "namespace": cls.Meta.namespace,
            },
        ),
    )
    return make_dataclass("Body", [input_field])


@dataclass
class Header:
    class Meta:
        namespace = "http://www.w3.org/2003/05/soap-envelope"

    action: Optional[str] = field(
        default=None,
        metadata={
            "name": "Action",
            "type": "Element",
            "namespace": "http://www.w3.org/2005/08/addressing",
        },
    )
    message_id: Optional[str] = field(
        default=None,
        metadata={
            "name": "MessageID",
            "type": "Element",
            "namespace": "http://www.w3.org/2005/08/addressing",
        },
    )
    relates_to: Optional[str] = field(
        default=None,
        metadata={
            "name": "RelatesTo",
            "type": "Element",
            "namespace": "http://www.w3.org/2005/08/addressing",
        },
    )


T = TypeVar("T")


def parse_action(content, input_cls: T) -> T:
    context = XmlContext()
    parser = UserXmlParser(context=context)
    result: input_cls = parser.from_string(content, input_cls)
    return result


def soap_input_factory(field_name, soap_namespace, cls_name, cls):
    body_cls = soap_body_factory(field_name, cls)
    fields: Iterable[str, type, Field] = [
        (
            "header",
            Optional[Header],
            field(
                default=None,
                metadata=dict(name="Header", type="Element", namespace=soap_namespace),
            ),
        ),
        (
            "body",
            Optional[body_cls],
            field(
                default=None,
                metadata=dict(name="Body", type="Element", namespace=soap_namespace),
            ),
        ),
    ]

    class Meta:
        name = "Envelope"
        namespace = soap_namespace

    return (
        make_dataclass(
            cls_name=cls_name,
            fields=fields,
            bases=(Envelope,),
            namespace={"Meta": Meta},
        ),
        body_cls,
    )


class SoapService:
    transport = "http://schemas.xmlsoap.org/soap/http"
    input_request = None
    output_response = None
    style = "document"
    soap_namespace = "http://www.w3.org/2003/05/soap-envelope"

    def __init_subclass__(cls, scm_type=None, name=None, **kwargs):
        cls.input, cls.input_body_class = soap_input_factory(
            "input", cls.soap_namespace, f"{cls.__name__}Input", cls.input_request
        )
        cls.output, cls.output_body_class = soap_output_factory(
            "output", cls.soap_namespace, f"{cls.__name__}Output", cls.output_response
        )



from multipart.multipart import parse_options_header
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

from starlette.endpoints import HTTPEndpoint
from starlette.responses import Response
from teserak.w3c.soap.types import SoapRequest, SoapRequestInput
from teserak.w3c.soap.utils import parse_action


def build_soap_request(input_cls, request_input):
    result = parse_action(request_input.body, input_cls)
    soap_request = SoapRequest[input_cls]()
    soap_request.input = result.body.input
    return soap_request


async def call_action(service, request_input: SoapRequestInput):
    action = service.get_action(request_input.action_name)

    input_cls = action.input
    output_cls = action.output
    action_handler = action.handler

    soap_request = build_soap_request(input_cls, request_input)
    result = await action_handler(soap_request)
    return output_cls(body=action.output_body_class(result))


async def render_response(response_instance):
    config = SerializerConfig(pretty_print=True)
    serializer = XmlSerializer(config=config)
    response_content = serializer.render(response_instance)
    return response_content


class SoapParser:
    namespace = None

    def get_action_name(self, headers, options) -> str:
        ...

    async def get_body(self, stream) -> str:
        body = b""
        async for chunk in stream:
            body += chunk
        return body.decode("utf-8")

    async def parse(self, headers: dict, stream):
        content_type_header = headers["content-type"]
        content_type, options = parse_options_header(content_type_header)
        action_name = self.get_action_name(headers, options)
        if b"boundary" in options:
            raise NotImplementedError()
        else:
            body = await self.get_body(stream)
            parts = {}
        return SoapRequestInput(action_name=action_name, body=body, parts=parts)


class SoapParser11(SoapParser):
    namespace = "http://schemas.xmlsoap.org/soap/envelope/"

    def get_action_name(self, headers, options):
        return headers["soapaction"].strip('"')


class SoapParser12(SoapParser):
    namespace = "http://www.w3.org/2003/05/soap-envelope"

    def get_action_name(self, headers, options):
        return options[b"action"].decode("utf-8")


class XmlResponse(Response):
    media_type = "application/xml"


def get_soap_parser(headers: dict):
    if "soapaction" in headers:
        parser = SoapParser11()
    else:
        parser = SoapParser12()
    return parser


async def parse_input(headers: dict, stream):
    parser = get_soap_parser(headers)

    return await parser.parse(headers, stream)


class SoapHandler(HTTPEndpoint):
    service = None

    async def post(self, request):
        soap_request_input = await parse_input(request.headers, request.stream())
        response_instance = await call_action(self.service, soap_request_input)
        response_content = await render_response(response_instance)
        return XmlResponse(response_content)



import typing
from dataclasses import dataclass, field

R = typing.TypeVar("R")


class SoapRequest(typing.Generic[R]):
    input: R

    def get_input_type(cls):
        return cls.__orig_bases__[0]


@dataclass
class SoapRequestInput:
    action_name: str
    body: str
    parts: typing.Dict[str, typing.Any] = field(default_factory=dict)


@dataclass
class MultipartResponse:
    body: str
    parts: typing.Dict[str, typing.Any]



from setuptools import setup, find_packages


setup(
    name="teserak-soap-server",
    python_requires=">=3.8",
    version='0.0.3',
    url="https://github.com/teserak/teserak-soap-server",
    license="BSD",
    description="The ASGI SOAP Server.",
    long_description="The ASGI SOAP Server.",
    author="Konrad Rymczak",
    author_email="me@teserak.com",
    packages=find_packages(exclude=["tests*"]),
    package_data={"teserak": ["py.typed"]},
    include_package_data=True,
    install_requires=[
        "starlette",
        "xsdata",
        "python-multipart",
    ],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    zip_safe=False,
)
